@@ -40,6 +40,9 @@
 #include <unistd.h>
 #include <ctype.h>
 #include <locale.h>
+#include <fcntl.h>
+#include <utime.h>
+#include <limits.h>
 
 #ifdef HAVE_POPT
 #  include <popt.h>
@@ -78,13 +81,16 @@
 #define __unused__
 #endif
 
-int  glob_debug = -1;
+static int  glob_debug = -1;
 char glob_cancel = 0;
-int  glob_frames = 0;
-int  glob_interval = 0;
+static int  glob_frames = 0;
+static int  glob_interval = 0;
 
 GPParams gp_params;
 
+/* flag for SIGUSR1 handler */
+static volatile int capture_now = 0;
+
 
 /*! \brief Copy string almost like strncpy, converting to lower case.
  *
@@ -363,10 +369,14 @@ get_path_for_file (const char *folder, CameraFile *file, char **path)
 
 
 int
-save_camera_file_to_file (const char *folder, CameraFile *file)
-{
+save_camera_file_to_file (
+	const char *folder, CameraFile *file, const char *curname
+) {
 	char *path = NULL, s[1024], c[1024];
 	CameraFileType type;
+	int res;
+	time_t mtime;
+	struct utimbuf u;
 
 	CR (gp_file_get_type (file, &type));
 
@@ -412,9 +422,18 @@ save_camera_file_to_file (const char *folder, CameraFile *file)
                 printf (_("Saving file as %s\n"), s);
 		fflush (stdout);
         }
-	CR (gp_file_save (file, s));
+	if (curname) {
+		unlink(s);
+		if (-1 == rename (curname, s))
+			perror("rename");
+	}
+	res = gp_file_get_mtime (file, &mtime);
+        if ((res == GP_OK) && (mtime)) {
+                u.actime = mtime;
+                u.modtime = mtime;
+                utime (s, &u);
+        }
 	gp_params_run_hook(&gp_params, "download", s);
-
 	return (GP_OK);
 }
 
@@ -446,8 +465,9 @@ save_file_to_file (Camera *camera, GPContext *context, Flags flags,
 		   const char *folder, const char *filename,
 		   CameraFileType type)
 {
-        int res;
+        int fd, res;
         CameraFile *file;
+	char	tmpname[20], *tmpfilename;
 
 	if (flags & FLAGS_NEW) {
 		CameraFileInfo info;
@@ -476,10 +496,27 @@ save_file_to_file (Camera *camera, GPContext *context, Flags flags,
 			return (GP_ERROR_NOT_SUPPORTED);
 		}
 	}
-	
-        CR (gp_file_new (&file));
-        CR (gp_camera_file_get (camera, folder, filename, type,
-				file, context));
+	strcpy (tmpname, "tmpfileXXXXXX");
+	fd = mkstemp(tmpname);
+	if (fd == -1) {
+        	CR (gp_file_new (&file));
+		tmpfilename = NULL;
+	} else {
+		res = gp_file_new_from_fd (&file, fd);
+		if (res < GP_OK) {
+			close (fd);
+			unlink (tmpname);
+			return res;
+		}
+		tmpfilename = tmpname;
+	}
+        res = gp_camera_file_get (camera, folder, filename, type,
+				  file, context);
+	if (res < GP_OK) {
+		gp_file_unref (file);
+		if (tmpfilename) unlink (tmpfilename);
+		return res;
+	}
 
 	if (flags & FLAGS_STDOUT) {
                 const char *data;
@@ -487,17 +524,17 @@ save_file_to_file (Camera *camera, GPContext *context, Flags flags,
 
                 CR (gp_file_get_data_and_size (file, &data, &size));
 
-		if (flags & FLAGS_STDOUT_SIZE)
+		if (flags & FLAGS_STDOUT_SIZE) /* this will be difficult in fd mode */
                         printf ("%li\n", size);
                 fwrite (data, sizeof(char), size, stdout);
                 gp_file_unref (file);
+		unlink (tmpname);
                 return (GP_OK);
         }
-
-        res = save_camera_file_to_file (folder, file);
-
+        res = save_camera_file_to_file (folder, file, tmpfilename);
         gp_file_unref (file);
-
+	if ((res!=GP_OK) && tmpfilename)
+		unlink (tmpfilename);
         return (res);
 }
 
@@ -546,23 +583,35 @@ get_file_common (const char *arg, CameraFileType type )
 	return (GP_OK);
 }
 
+static void
+sig_handler_capture_now (int sig_num)
+{
+	signal (SIGUSR1, sig_handler_capture_now);
+	capture_now = 1;
+}
 
 int
-capture_generic (CameraCaptureType type, const char __unused__ *name)
+capture_generic (CameraCaptureType type, const char __unused__ *name, int download)
 {
 	CameraFilePath path, last;
 	char *pathsep;
 	int result, frames = 0;
-	time_t next_pic_time;
+	time_t next_pic_time, now;
 	int waittime;
 
 	next_pic_time = time (NULL) + glob_interval;
 	if(glob_interval) {
 		memset(&last, 0, sizeof(last));
-		if (!(gp_params.flags & FLAGS_QUIET))
-			printf (_("Time-lapse mode enabled (interval: %ds).\n"),
-				glob_interval);
+		if (!(gp_params.flags & FLAGS_QUIET)) {
+			if (glob_interval != -1)
+				printf (_("Time-lapse mode enabled (interval: %ds).\n"),
+					glob_interval);
+			else
+				printf (_("Standing by waiting for SIGUSR1 to capture.\n"));
+		}
 	}
+	capture_now = 0;
+	signal(SIGUSR1, sig_handler_capture_now);
 
 	while(++frames) {
 		if (!(gp_params.flags & FLAGS_QUIET) && glob_interval) {
@@ -602,63 +651,46 @@ capture_generic (CameraCaptureType type, const char __unused__ *name)
 			printf (_("New file is in location %s%s%s on the camera\n"),
 				path.folder, pathsep, path.name);
 
-		if(!glob_interval) break;
+		if (download) {
+			if (strcmp(path.folder, last.folder)) {
+				memcpy(&last, &path, sizeof(last));
 
-		if(strcmp(path.folder, last.folder)) {
-			memcpy(&last, &path, sizeof(last));
+				result = set_folder_action(&gp_params, path.folder);
+				if (result != GP_OK) {
+					cli_error_print(_("Could not set folder."));
+					return (result);
+				}
+			}
 
-			result = set_folder_action(&gp_params, path.folder);
+			result = get_file_common (path.name, GP_FILE_TYPE_NORMAL);
 			if (result != GP_OK) {
-				cli_error_print(_("Could not set folder."));
+				cli_error_print (_("Could not get image."));
+				if(result == GP_ERROR_FILE_NOT_FOUND) {
+					/* Buggy libcanon.so?
+					 * Can happen if this was the first capture after a
+					 * CF card format, or during a directory roll-over,
+					 * ie: CANON100 -> CANON101
+					 */
+					cli_error_print ( _("Buggy libcanon.so?"));
+				}
+				return (result);
+			}
+
+			if (!(gp_params.flags & FLAGS_QUIET))
+				printf (_("Deleting file %s%s%s on the camera\n"),
+					path.folder, pathsep, path.name);
+
+			result = delete_file_action (&gp_params, path.name);
+			if (result != GP_OK) {
+				cli_error_print ( _("Could not delete image."));
 				return (result);
 			}
 		}
-
-		/* XXX: I wonder if there is some way to determine if the camera
-		 * is still writing the image because if we don't sleep here,
-		 * the image will be corrupt when we attempt to download it.
-		 * I assume this is a result of the camera still writing the
-		 * data!  I picked 2 seconds as a first try and it seems like
-		 * enough time for high-quality JPEG on the Canon EOS 10D (with
-		 * a San Disk Ultra II CF).  May need to be increased for RAW
-		 * images...  I don't think this is the correct way to do this.
-		 */
-
-		/* Marcus Meissner: Fix the camera driver to only return from
-		 * camera_capture() when you know it is done ...
-		 * Otherwise other drivers suffer.
-		 */
-#if 0
-		sleep(2);
-#endif
-
-		result = get_file_common (path.name, GP_FILE_TYPE_NORMAL);
-		if (result != GP_OK) {
-			cli_error_print (_("Could not get image."));
-			if(result == GP_ERROR_FILE_NOT_FOUND) {
-				/* Buggy libcanon.so?
-				 * Can happen if this was the first capture after a
-				 * CF card format, or during a directory roll-over,
-				 * ie: CANON100 -> CANON101
-				 */
-				cli_error_print ( _("Buggy libcanon.so?"));
-			}
-			return (result);
-		}
-
-		if (!(gp_params.flags & FLAGS_QUIET))
-			printf (_("Deleting file %s%s%s on the camera\n"),
-				path.folder, pathsep, path.name);
-
-		result = delete_file_action (&gp_params, path.name);
-		if (result != GP_OK) {
-			cli_error_print ( _("Could not delete image."));
-			return (result);
-		}
-
 		/* Break if we've reached the requested number of frames
 		 * to capture.
 		 */
+		if(!glob_interval) break;
+
 		if(glob_frames && frames == glob_frames) break;
 
 #if 0
@@ -675,21 +707,129 @@ capture_generic (CameraCaptureType type, const char __unused__ *name)
 			cli_error_print (_("Could not close camera connection."));
 		}
 #endif
-		waittime = next_pic_time - time (NULL);
-		if (waittime > 0) {
-			if (!(gp_params.flags & FLAGS_QUIET) && glob_interval)
-				printf (_("Sleeping for %d second(s)...\n"), glob_interval);
-			sleep (waittime);
+		/*
+		 * Even if the interval is set to -1, it is better to take a
+		 * picture first to prepare the camera driver for faster
+		 * response when a signal is caught.
+		 * [alesan]
+		 */
+		if (glob_interval != -1) {
+			waittime = next_pic_time - time (NULL);
+			if (waittime > 0) {
+				int i;
+				if (!(gp_params.flags & FLAGS_QUIET) && glob_interval)
+					printf (_("Sleeping for %d second(s)...\n"), waittime);
+				/* We're not sure about sleep() semantics when catching a signal */
+				for (i=0; (!capture_now) && (i<waittime); i++)
+					sleep(1);
+				if (capture_now && !(gp_params.flags & FLAGS_QUIET) && glob_interval)
+					printf (_("Awakened by SIGUSR1...\n"));
+			} else {
+				if (!(gp_params.flags & FLAGS_QUIET) && glob_interval)
+					printf (_("not sleeping (%d seconds behind schedule)\n"), - waittime);
+			}
+			if (capture_now && (gp_params.flags & FLAGS_RESET_CAPTURE_INTERVAL))
+				next_pic_time = time(NULL) + glob_interval;
+			else if (!capture_now) {
+				now = time(NULL) - glob_interval;
+				/*
+				 * In the case of a (huge) time-sync while gphoto is running,
+				 * gphoto could percieve an extremely large amount of time and
+				 * stay "behind schedule" quite forever. That's why I reduce the
+				 * difference of time with the following loop.
+				 * [alesan]
+				 */
+				do {
+					next_pic_time += glob_interval;
+				} while (next_pic_time < now);
+			}
+			capture_now = 0;
 		} else {
-			if (!(gp_params.flags & FLAGS_QUIET) && glob_interval)
-				printf (_("not sleeping (%d seconds behind schedule)\n"), - waittime);
+			/* wait indefinitely for SIGUSR1 */
+			do {
+				pause();
+			} while(!capture_now);
+			capture_now = 0;
+			if (!(gp_params.flags & FLAGS_QUIET))
+				printf (_("Awakened by SIGUSR1...\n"));
 		}
-		next_pic_time += glob_interval;
 	}
 
+	signal(SIGUSR1, SIG_DFL);
 	return (GP_OK);
 }
 
+static int
+capture_tethered (const char __unused__ *name)
+{
+	CameraFilePath last;
+	int result;
+        CameraEventType event;
+        CameraFilePath  *fn;
+        void    *data = NULL;
+
+	printf ( _("Waiting for events from camera. Press Ctrl-C to abort.\n"));
+
+	while (1) {
+		result = gp_camera_wait_for_event (gp_params.camera, 10000, &event, &data, gp_params.context);
+
+		if (result != GP_OK)
+			break;
+		switch (event) {
+		case GP_EVENT_UNKNOWN:
+			if (data) {
+				printf("UNKNOWN %s\n", (char*)data);
+			} else {
+				printf("UNKNOWN\n");
+			}
+			break;
+		case GP_EVENT_TIMEOUT:
+			/* just continue and wait for next one. */
+			break;
+		case GP_EVENT_FILE_ADDED:
+			fn = (CameraFilePath*)data;
+			printf( _("New file %s/%s, downloading...\n"),fn->name, fn->folder);
+
+			if(strcmp(fn->folder, last.folder)) {
+				strcpy(last.folder, fn->folder);
+				result = set_folder_action(&gp_params, fn->folder);
+				if (result != GP_OK) {
+					cli_error_print(_("Could not set folder."));
+					return (result);
+				}
+			}
+			result = get_file_common (fn->name, GP_FILE_TYPE_NORMAL);
+			if (result != GP_OK) {
+				cli_error_print (_("Could not get image."));
+				if(result == GP_ERROR_FILE_NOT_FOUND) {
+					/* Buggy libcanon.so?
+					 * Can happen if this was the first capture after a
+					 * CF card format, or during a directory roll-over,
+					 * ie: CANON100 -> CANON101
+					 */
+					cli_error_print ( _("Buggy libcanon.so?"));
+				}
+				return (result);
+			}
+
+			do {
+				result = delete_file_action (&gp_params, fn->name);
+			} while (result == GP_ERROR_CAMERA_BUSY);
+			if (result != GP_OK) {
+				cli_error_print ( _("Could not delete image."));
+				/* continue in tethered loop */
+			}
+			break;
+		case GP_EVENT_FOLDER_ADDED:
+			fn = (CameraFilePath*)data;
+			/*printf("FOLDERADDED %s %s\n",fn->name, fn->folder);*/
+			break;
+        	}
+	}
+        return GP_OK;
+}
+
+
 
 /* Set/init global variables                                    */
 /* ------------------------------------------------------------ */
@@ -822,9 +962,11 @@ typedef enum {
 	ARG_CAPTURE_FRAMES,
 	ARG_CAPTURE_INTERVAL,
 	ARG_CAPTURE_IMAGE,
+	ARG_CAPTURE_IMAGE_AND_DOWNLOAD,
 	ARG_CAPTURE_MOVIE,
 	ARG_CAPTURE_PREVIEW,
 	ARG_CAPTURE_SOUND,
+	ARG_CAPTURE_TETHERED,
 	ARG_CONFIG,
 	ARG_DEBUG,
 	ARG_DEBUG_LOGFILE,
@@ -846,6 +988,7 @@ typedef enum {
 	ARG_GET_RAW_DATA,
 	ARG_GET_THUMBNAIL,
 	ARG_HELP,
+	ARG_HOOK_SCRIPT,
 	ARG_LIST_CAMERAS,
 	ARG_LIST_CONFIG,
 	ARG_LIST_FILES,
@@ -860,6 +1003,7 @@ typedef enum {
 	ARG_PORT,
 	ARG_QUIET,
 	ARG_RECURSE,
+	ARG_RESET_INTERVAL,
 	ARG_RMDIR,
 	ARG_SHELL,
 	ARG_SHOW_EXIF,
@@ -867,14 +1011,14 @@ typedef enum {
 	ARG_SPEED,
 	ARG_STDOUT,
 	ARG_STDOUT_SIZE,
+	ARG_STORAGE_INFO,
 	ARG_SUMMARY,
 	ARG_UPLOAD_FILE,
 	ARG_UPLOAD_METADATA,
 	ARG_USAGE,
 	ARG_USBID,
 	ARG_VERSION,
-	ARG_WAIT_EVENT,
-	ARG_HOOK_SCRIPT,
+	ARG_WAIT_EVENT
 } Arg;
 
 typedef enum {
@@ -1008,6 +1152,10 @@ cb_arg_init (poptContext __unused__ ctx,
 		gp_params.flags |= FLAGS_QUIET;
 		break;
 
+	case ARG_RESET_INTERVAL:
+		gp_params.flags |= FLAGS_RESET_CAPTURE_INTERVAL;
+		break;
+
 	case ARG_HOOK_SCRIPT:
 		do {
 			const size_t sz = strlen(arg);
@@ -1019,7 +1167,7 @@ cb_arg_init (poptContext __unused__ ctx,
 			gp_params.hook_script = strcpy(copy, arg);
 			/* Run init hook */
 			if (0!=gp_params_run_hook(&gp_params, "init", NULL)) {
-				fprintf(stdout,
+				fprintf(stderr,
 					"Hook script \"%s\" init failed. Aborting.\n",
 					gp_params.hook_script);
 				exit(3);
@@ -1072,16 +1220,22 @@ cb_arg_run (poptContext __unused__ ctx,
 		params->p.r = auto_detect_action (&gp_params);
 		break;
 	case ARG_CAPTURE_IMAGE:
-		params->p.r = capture_generic (GP_CAPTURE_IMAGE, arg);
+		params->p.r = capture_generic (GP_CAPTURE_IMAGE, arg, 0);
+		break;
+	case ARG_CAPTURE_IMAGE_AND_DOWNLOAD:
+		params->p.r = capture_generic (GP_CAPTURE_IMAGE, arg, 1);
 		break;
 	case ARG_CAPTURE_MOVIE:
-		params->p.r = capture_generic (GP_CAPTURE_MOVIE, arg);
+		params->p.r = capture_generic (GP_CAPTURE_MOVIE, arg, 0);
 		break;
 	case ARG_CAPTURE_PREVIEW:
 		params->p.r = action_camera_capture_preview (&gp_params);
 		break;
 	case ARG_CAPTURE_SOUND:
-		params->p.r = capture_generic (GP_CAPTURE_SOUND, arg);
+		params->p.r = capture_generic (GP_CAPTURE_SOUND, arg, 0);
+		break;
+	case ARG_CAPTURE_TETHERED:
+		params->p.r = capture_tethered (arg);
 		break;
 	case ARG_CONFIG:
 #ifdef HAVE_CDK
@@ -1224,6 +1378,9 @@ cb_arg_run (poptContext __unused__ ctx,
 	case ARG_WAIT_EVENT:
 		params->p.r = action_camera_wait_event (&gp_params);
 		break;
+	case ARG_STORAGE_INFO:
+		params->p.r = print_storage_info (&gp_params);
+		break;
 	default:
 		break;
 	};
@@ -1297,7 +1454,7 @@ report_failure (int result, int argc, char **argv)
 		 */
 		printf ("    env LANG=C gphoto2 --debug --debug-logfile=my-logfile.txt");
 		for (n = 1; n < argc; n++) {
-			if (argv[n][0] == '-')
+			if (strchr(argv[n], ' ') == NULL)
 				printf(" %s",argv[n]);
 			else
 				printf(" \"%s\"",argv[n]);
@@ -1433,12 +1590,18 @@ main (int argc, char **argv, char **envp)
 		 N_("Set number of frames to capture (default=infinite)"), N_("COUNT")},
 		{"interval", 'I', POPT_ARG_INT, NULL, ARG_CAPTURE_INTERVAL,
 		 N_("Set capture interval in seconds"), N_("SECONDS")},
+		{"reset-interval", '\0', POPT_ARG_NONE, NULL, ARG_RESET_INTERVAL,
+		 N_("Reset capture interval on signal (default=no)"), NULL},
 		{"capture-image", '\0', POPT_ARG_NONE, NULL,
 		 ARG_CAPTURE_IMAGE, N_("Capture an image"), NULL},
+		{"capture-image-and-download", '\0', POPT_ARG_NONE, NULL,
+		 ARG_CAPTURE_IMAGE_AND_DOWNLOAD, N_("Capture an image and download it"), NULL},
 		{"capture-movie", '\0', POPT_ARG_NONE, NULL,
 		 ARG_CAPTURE_MOVIE, N_("Capture a movie"), NULL},
 		{"capture-sound", '\0', POPT_ARG_NONE, NULL,
 		 ARG_CAPTURE_SOUND, N_("Capture an audio clip"), NULL},
+		{"capture-tethered", '\0', POPT_ARG_NONE, NULL,
+		 ARG_CAPTURE_TETHERED, N_("Wait for shutter release on the camera and download"), NULL},
 		POPT_TABLEEND
 	};
 	const struct poptOption fileOptions[] = {
@@ -1521,6 +1684,8 @@ main (int argc, char **argv, char **envp)
 		 N_("Show camera driver manual"), NULL},
 		{"about", '\0', POPT_ARG_NONE, NULL, ARG_ABOUT,
 		 N_("About the camera driver manual"), NULL},
+		{"storage-info", '\0', POPT_ARG_NONE, NULL, ARG_STORAGE_INFO,
+		 N_("Show storage information"), NULL},
 		{"shell", '\0', POPT_ARG_NONE, NULL, ARG_SHELL,
 		 N_("gPhoto shell"), NULL},
 		POPT_TABLEEND
@@ -1566,6 +1731,8 @@ main (int argc, char **argv, char **envp)
 	ctx = poptGetContext (PACKAGE, argc, (const char **) argv, options, 0);
 	if (argc <= 1) {
 		poptPrintUsage (ctx, stdout, 0);
+		gp_params_exit (&gp_params);
+                poptFreeContext(ctx);
 		return (0);
 	}
 
@@ -1577,14 +1744,20 @@ main (int argc, char **argv, char **envp)
 	while ((result = poptGetNextOpt (ctx)) >= 0);
 	if (result == POPT_ERROR_BADOPT) {
 		poptPrintUsage (ctx, stderr, 0);
+		gp_params_exit (&gp_params);
+                poptFreeContext(ctx);
 		return (EXIT_FAILURE);
 	}
 	if (help_option_given) {
 		poptPrintHelp(ctx, stdout, 0);
+		gp_params_exit (&gp_params);
+                poptFreeContext(ctx);
 		return 0;
 	}
 	if (usage_option_given) {
 		poptPrintUsage(ctx, stdout, 0);
+		gp_params_exit (&gp_params);
+                poptFreeContext(ctx);
 		return 0;
 	}
 	if (debug_option_given) {
@@ -1603,6 +1776,28 @@ main (int argc, char **argv, char **envp)
 	cb_params.type = CALLBACK_PARAMS_TYPE_INITIALIZE;
 	poptResetContext (ctx);
 	while ((cb_params.p.r >= 0) && (poptGetNextOpt (ctx) >= 0));
+	/* Load default values for --filename and --hook-script if not
+	 * explicitely specified
+	 */
+	if (!gp_params.filename) {
+		char buf[256];
+		if (gp_setting_get("gphoto2","filename",buf)>=0) {
+			set_filename_action(&gp_params,buf);
+		}	
+	}
+	if (!gp_params.hook_script) {
+		char buf[PATH_MAX];
+		if (gp_setting_get("gphoto2","hook-script",buf)>=0) {
+			gp_params.hook_script = strdup(buf);
+			/* Run init hook */
+			if (0!=gp_params_run_hook(&gp_params, "init", NULL)) {
+				fprintf(stdout,
+					"Hook script \"%s\" init failed. Aborting.\n",
+					gp_params.hook_script);
+				exit(3);
+			}
+		}	
+	}
 	CR_MAIN (cb_params.p.r);
 
 #define CHECK_OPT(o)					\
@@ -1624,6 +1819,7 @@ main (int argc, char **argv, char **envp)
 	CHECK_OPT (ARG_CAPTURE_MOVIE);
 	CHECK_OPT (ARG_CAPTURE_PREVIEW);
 	CHECK_OPT (ARG_CAPTURE_SOUND);
+	CHECK_OPT (ARG_CAPTURE_TETHERED);
 	CHECK_OPT (ARG_CONFIG);
 	CHECK_OPT (ARG_DELETE_ALL_FILES);
 	CHECK_OPT (ARG_DELETE_FILE);
@@ -1646,6 +1842,7 @@ main (int argc, char **argv, char **envp)
 	CHECK_OPT (ARG_SHOW_INFO);
 	CHECK_OPT (ARG_SUMMARY);
 	CHECK_OPT (ARG_UPLOAD_FILE);
+	CHECK_OPT (ARG_STORAGE_INFO);
 	if (cb_params.p.q.found &&
 	    (!strcmp (a.model, "") || (info.type == GP_PORT_NONE))) {
 		int count;
@@ -1832,7 +2029,7 @@ main (int argc, char **argv, char **envp)
 	 *        we load the camlibs */
 
 	gp_params_exit (&gp_params);
-
+        poptFreeContext(ctx);
         return (EXIT_SUCCESS);
 }
 
